/***************************************************************************/
/*  Program to do a basic simulation and analysis of IBD inference with LD */
/*  Routines for forwards and back computation:  December 2008             */
/* Contains   actual_freq():  computes frequencies from real data          */
/*            make_freq():    puts frequencies into freq-array             */
/*            hap_set_penmat():  set penetrance matrices for analysis      */
/*            set_marg_ibd():  sets marginal ibd of states                 */
/*            set_transmat():  sets state transition matrix                */
/*            hap_fwd_compute():  does forward computation                 */
/*            hap_bck_compute():  does backward computation                */
/*            my_set_up():    sets up MORGAN framework                     */
/*                                                                         */
/***************************************************************************/

void set_tmat (int ns)
{
   if (ns == 15)
   {
      tmat[0][2] = tmat[0][3] = tmat[0][5] = tmat[0][6] = hh;
      tmat[1][4] = tmat[1][7] = hh;
      tmat[8][10] = tmat[8][13] = tmat[9][11] = tmat[9][12] = hh;
      tmat[0][0] = -4 * hh;
      tmat[1][1] = tmat[8][8] = tmat[9][9] = -2 * hh;

      tmat[2][0] = tmat[3][0] = tmat[5][0] = tmat[6][0] = 3 * gg;
      tmat[2][2] = tmat[3][3] = tmat[5][5] = tmat[6][6] = -3 * (gg + hh);
      tmat[2][4] = tmat[3][4] = tmat[5][7] = tmat[6][8] = hh;
      tmat[2][10] = tmat[2][12] = tmat[3][11] = tmat[3][13] = hh;
      tmat[5][10] = tmat[5][11] = tmat[6][12] = tmat[6][13] = hh;

      tmat[4][14] = tmat[7][14] = hh;
      tmat[4][1] = tmat[7][1] = gg;
      tmat[4][2] = tmat[4][3] = tmat[7][5] = tmat[7][6] = 2 * gg;
      tmat[4][4] = tmat[7][7] = -5 * gg - hh;

      tmat[10][14] = tmat[11][14] = tmat[12][14] = tmat[13][14] = hh;
      tmat[10][8] = tmat[11][9] = tmat[12][9] = tmat[13][8] = gg;
      tmat[10][2] = tmat[11][3] = tmat[12][2] = tmat[13][2] = 2 * gg;
      tmat[10][5] = tmat[11][5] = tmat[12][6] = tmat[13][6] = 2 * gg;
      tmat[10][10] = tmat[11][11] = tmat[12][12] = tmat[13][13] = -5 * gg - hh;

      tmat[14][10] = tmat[14][11] = tmat[14][12] = tmat[14][13] = gg;
      tmat[14][4] = tmat[14][7] = gg;
      tmat[14][14] = -6 * gg;
   }
   else if (ns == 9)
   {

      tmat[0][2] = tmat[0][4] = 2.0 * hh;
      tmat[1][3] = tmat[1][5] = hh;
      tmat[6][7] = 2.0 * hh;
      tmat[0][0] = -4 * hh;
      tmat[1][1] = tmat[6][6] = -2 * hh;

      tmat[2][0] = tmat[4][0] = 3 * gg;
      tmat[2][2] = tmat[4][4] = -3 * (gg + hh);
      tmat[2][3] = tmat[4][5] = hh;
      tmat[2][7] = tmat[4][7] = 2.0 * hh;

      tmat[3][8] = tmat[5][8] = hh;
      tmat[3][1] = tmat[5][1] = gg;
      tmat[3][2] = tmat[5][4] = 4 * gg;
      tmat[3][3] = tmat[5][5] = -5 * gg - hh;

      tmat[7][8] = hh;
      tmat[7][6] = gg;
      tmat[7][2] = tmat[7][4] = 2 * gg;
      tmat[7][7] = -5 * gg - hh;

      tmat[8][7] = 4.0 * gg;
      tmat[8][5] = tmat[8][3] = gg;
      tmat[8][8] = -6 * gg;
   }
   return;
}

void hap_set_penmat (double p0, double *ppvec, int mph, int ms, int phdat,
                     double errate, int igdx, int phased)
/* set the penetrance vector *ppvec for data phdat */
/* two haps; mph=, ms =2  */
/* 2 genos;  mph=9, ms =9 */
/* 4 haps;  mph=16, ms = 15 */
/* THIS IS REWRITTEN -- ONLY SET FOR DATA IN QUESTION
   Set a vector over the states, for the gives data at locus j
   Allow for missing data: phdat = -1                          */
{
   double p1, pg0, pg1, pg2;
   int   j;

   if (phdat == -1)
   {
      for (j = 0; j < ms; ++j)
         ppvec[j] = 1.0;
      return;
   }

   p1 = 1.0 - p0;

   if ((ms == 2) && (mph == 4))
   {
      /* ss=1; non-ibd state -- NOTE CHANGE  */
      /* ss=0 ; ibd states; -- NOTE CHANGE   */
      if (phdat == 0)
      {
         ppvec[1] = p0 * p0;
         ppvec[0] = p0 * (1.0 - errate) + errate * ppvec[1];
         return;
      }
      if ((phdat == 1) || (phdat == 2))
      {
         ppvec[1] = p0 * p1;
         ppvec[0] = errate * ppvec[1];
         return;
      }
      if (phdat == 3)
      {
         ppvec[1] = p1 * p1;
         ppvec[0] = p1 * (1.0 - errate) + errate * ppvec[1];
         return;
      }
   }
   else if ((ms == 9) && (mph == 9))
   {
      pg0 = p0 * p0;
      pg1 = 2 * p0 * p1;
      pg2 = p1 * p1;

      for (j = 0; j < ms; ++j)
         ppvec[j] = 0.0;

      if (phdat == 0)
      {
         ppvec[8] = pg0 * pg0;
         ppvec[0] = p0;
         ppvec[1] = pg0;
         ppvec[2] = pg0;
         ppvec[3] = pg0;
         ppvec[6] = pg0;
         ppvec[4] = ppvec[5] = ppvec[7] = p0 * p0 * p0;
      }
      else if (phdat == 1)
      {
         ppvec[8] = pg0 * pg1;
         ppvec[2] = p0 * p1;
         ppvec[4] = pg1 * p0;
         ppvec[7] = pg0 * p1;
      }
      else if (phdat == 2)
      {
         ppvec[8] = pg0 * pg2;
         ppvec[1] = p0 * p1;
         ppvec[4] = pg2 * p0;
         ppvec[5] = pg0 * p1;
      }
      else if (phdat == 3)
      {
         ppvec[8] = pg1 * pg0;
         ppvec[3] = p0 * p1;
         ppvec[5] = pg1 * p0;
         ppvec[7] = pg0 * p1;
      }
      else if (phdat == 4)
      {
         ppvec[8] = pg1 * pg1;
         ppvec[6] = pg1;
         ppvec[7] = p0 * p1;
      }
      else if (phdat == 5)
      {
         ppvec[8] = pg1 * pg2;
         ppvec[3] = p0 * p1;
         ppvec[5] = pg1 * p1;
         ppvec[7] = pg2 * p0;
      }
      else if (phdat == 6)
      {
         ppvec[8] = pg2 * pg0;
         ppvec[1] = p0 * p1;
         ppvec[4] = pg0 * p1;
         ppvec[5] = pg2 * p0;
      }
      else if (phdat == 7)
      {
         ppvec[8] = pg2 * pg1;
         ppvec[2] = p0 * p1;
         ppvec[4] = pg1 * p1;
         ppvec[7] = pg2 * p0;
      }
      else if (phdat == 8)
      {
         ppvec[8] = pg2 * pg2;
         ppvec[0] = p1;
         ppvec[1] = pg2;
         ppvec[2] = pg2;
         ppvec[3] = pg2;
         ppvec[6] = pg2;
         ppvec[4] = ppvec[5] = ppvec[7] = p1 * p1 * p1;
      }
   }

   else if ((ms == 15) && (mph == 16))
   {
      if ((igdx == 0) || (igdx == 2 && phased))
      {
         /* phenos in binary 0000 to 1111
            states 1111,1122,1112,1121,1123, 1211,1222,1233,
            1212,1221,  1213,  1231, 1223, 1232, 1234  */
         for (j = 0; j < ms; ++j)
         {
            if (phdat < 8)
               ppvec[j] = p0;
            else
               ppvec[j] = p1;
         }

         for (j = 0; j < 5; ++j)        /* first 2 genes ibd */
            if ((phdat > 3) && (phdat < 12))
               ppvec[j] = 0.0;          /* a1 != a2 */

         for (j = 5; j < ms; ++j)       /* first two genes not ibd */
         {
            if (phdat < 4)
               ppvec[j] *= p0;
            if ((phdat > 3) && (phdat < 8))
               ppvec[j] *= p1;
            if ((phdat > 7) && (phdat < 12))
               ppvec[j] *= p0;
            if (phdat > 11)
               ppvec[j] *= p1;
         }

         if ((phdat > 0) && (phdat < 15))
            ppvec[0] = 0.0;             /* all genes ibd state 0 */
         /* state 1122 */
         if ((phdat == 1) || (phdat == 2) || (phdat == 13) || (phdat == 14))
            ppvec[1] = 0.0;

         if ((phdat == 0) || (phdat == 12))
            ppvec[1] *= p0;

         if ((phdat == 3) || (phdat == 15))
            ppvec[1] *= p1;

         if ((phdat > 1) && (phdat < 14)) /* first 3 genes ibd */
            ppvec[2] = 0.0;

         if ((phdat == 0) || (phdat == 14))
            ppvec[2] *= p0;

         if ((phdat == 1) || (phdat == 15))
            ppvec[2] *= p1;

         /* state 1121 */
         if ((phdat == 0) || (phdat == 13))
            ppvec[3] *= p0;
         else if ((phdat == 2) || (phdat == 15))
            ppvec[3] *= p1;
         else
            ppvec[3] = 0.0;

         if (phdat < 4)                 /* state 1121, B1 is 00 */
         {
            if ((phdat == 0) || (phdat == 1))
               ppvec[4] *= p0;
            else
               ppvec[4] *= p1;
            if ((phdat == 0) || (phdat == 2))
               ppvec[4] *= p0;
            else
               ppvec[4] *= p1;
         }
         if (phdat > 11)                /* state 1121, B1 is 11 */
         {
            if ((phdat == 12) || (phdat == 13))
               ppvec[4] *= p0;
            else
               ppvec[4] *= p1;
            if ((phdat == 12) || (phdat == 14))
               ppvec[4] *= p0;
            else
               ppvec[4] *= p1;
         }

         /* now do last 2 genes states 5 through 15  */
         /*   states 5,6,7; last 2 genes are ibd     */
         if ((phdat == 0) || (phdat == 4) || (phdat == 8) || (phdat == 12))
            ppvec[7] *= p0;
         else if ((phdat == 3) || (phdat == 7) || (phdat == 11)
                  || (phdat == 15))
            ppvec[7] *= p1;
         else
            ppvec[7] = 0.0;

         if ((phdat == 0) || (phdat == 4) || (phdat == 11) || (phdat == 15))
            ppvec[5] *= 1;
         else
            ppvec[5] = 0.0;

         if ((phdat == 0) || (phdat == 7) || (phdat == 8) || (phdat == 15))
            ppvec[6] *= 1;
         else
            ppvec[6] = 0.0;

         /* states 8,9: 1212 and 1221: only rules out */
         if (((phdat > 0) && (phdat < 5)) || (phdat == 7) || (phdat == 8)
             || ((phdat > 10) && (phdat < 15)))
         {
            ppvec[8] = 0.0;
            ppvec[9] = 0.0;
         }
         if ((phdat == 5) || (phdat == 10))
            ppvec[9] = 0.0;
         if ((phdat == 6) || (phdat == 9))
            ppvec[8] = 0.0;

         /* states 10,11,12,13: 1213,  1231, 1223, 1232 */
         for (j = 10; j < 14; ++j)
         {
            if (phdat == 0)
               ppvec[j] *= p0;
            if (phdat == 1)
            {
               if ((j == 10) || (j == 12))
                  ppvec[j] *= p1;
               else
                  ppvec[j] = 0.0;
            }
            if (phdat == 2)
            {
               if ((j == 11) || (j == 13))
                  ppvec[j] *= p1;
               else
                  ppvec[j] = 0.0;
            }
            if ((phdat == 3) || (phdat == 12))
               ppvec[j] = 0.0;
            if (phdat == 4)
            {
               if ((j == 11) || (j == 10))
                  ppvec[j] *= p0;
               else
                  ppvec[j] = 0.0;
            }
            if (phdat == 7)
            {
               if ((j == 12) || (j == 13))
                  ppvec[j] *= p1;
               else
                  ppvec[j] = 0.0;
            }
            if (phdat == 5)
            {
               if (j == 10)
                  ppvec[j] *= p1;
               else if (j == 13)
                  ppvec[j] *= p0;
               else
                  ppvec[j] = 0.0;
            }
            if (phdat == 6)
            {
               if (j == 11)
                  ppvec[j] *= p1;
               else if (j == 12)
                  ppvec[j] *= p0;
               else
                  ppvec[j] = 0.0;
            }
            if (phdat == 8)
            {
               if ((j == 12) || (j == 13))
                  ppvec[j] *= p0;
               else
                  ppvec[j] = 0.0;
            }
            if (phdat == 11)
            {
               if ((j == 11) || (j == 10))
                  ppvec[j] *= p1;
               else
                  ppvec[j] = 0.0;
            }
            if (phdat == 9)
            {
               if (j == 11)
                  ppvec[j] *= p0;
               else if (j == 12)
                  ppvec[j] *= p1;
               else
                  ppvec[j] = 0.0;
            }
            if (phdat == 10)
            {
               if (j == 10)
                  ppvec[j] *= p0;
               else if (j == 13)
                  ppvec[j] *= p1;
               else
                  ppvec[j] = 0.0;
            }
            if (phdat == 14)
            {
               if ((j == 10) || (j == 12))
                  ppvec[j] *= p0;
               else
                  ppvec[j] = 0.0;
            }
            if (phdat == 13)
            {
               if ((j == 11) || (j == 13))
                  ppvec[j] *= p0;
               else
                  ppvec[j] = 0.0;
            }
            if (phdat == 15)
               ppvec[j] *= p1;
         }

         /* state 14:  no genes ibd  */
         if ((phdat == 0) || (phdat == 4) || (phdat == 8) || (phdat == 12))
            ppvec[14] *= (p0 * p0);
         else if ((phdat == 3) || (phdat == 7) || (phdat == 11)
                  || (phdat == 15))
            ppvec[14] *= (p1 * p1);
         else
            ppvec[14] *= (p0 * p1);
      }
      else if (igdx == 2 && phased == 0)
      {
         /* this is the penetrance matrix/emission probabilities
          * for when we have mixed haplo and geno data, and this
          * locus is unphased (i.e. genotypic data). The phenos
          * are expressed as haplotypic phenotypes, but the  
          * probabilities for different hidden states are
          * collapsed into the 9 distinct genotypic states */
          /**********************************************/
         pg0 = p0 * p0;
         pg1 = 2 * p0 * p1;
         pg2 = p1 * p1;

         for (j = 0; j < ms; ++j)
            ppvec[j] = 0.0;
         if (phdat == 0)
         {
            ppvec[14] = pg0 * pg0;
            ppvec[0] = p0;
            ppvec[1] = pg0;
            ppvec[2] = ppvec[3] = pg0;
            ppvec[4] = pg0;
            ppvec[8] = ppvec[9] = pg0;
            ppvec[5] = ppvec[6] = ppvec[7] = ppvec[10] = ppvec[11] = ppvec[12] =
               ppvec[13] = p0 * p0 * p0;
         }
         else if ((phdat == 1) || (phdat == 2))
         {
            ppvec[14] = pg0 * pg1;
            ppvec[2] = ppvec[3] = p0 * p1;
            ppvec[5] = ppvec[6] = pg1 * p0;
            ppvec[10] = ppvec[11] = ppvec[12] = ppvec[13] = pg0 * p1;
         }
         else if ((phdat == 3))
         {
            ppvec[14] = pg0 * pg2;
            ppvec[1] = p0 * p1;
            ppvec[5] = ppvec[6] = pg2 * p0;
            ppvec[7] = pg0 * p1;
         }
         else if ((phdat == 4) || (phdat == 8))
         {
            ppvec[14] = pg1 * pg0;
            ppvec[4] = p0 * p1;
            ppvec[7] = pg1 * p0;
            ppvec[10] = ppvec[11] = ppvec[12] = ppvec[13] = pg0 * p1;
         }
         else if ((phdat == 5) || (phdat == 6) || (phdat == 9) || (phdat == 10))
         {
            ppvec[14] = pg1 * pg1;
            ppvec[8] = ppvec[9] = pg1;
            ppvec[10] = ppvec[11] = ppvec[12] = ppvec[13] = p0 * p1;
         }
         else if ((phdat == 7) || (phdat == 11))
         {
            ppvec[14] = pg1 * pg2;
            ppvec[4] = p0 * p1;
            ppvec[7] = pg1 * p1;
            ppvec[10] = ppvec[11] = ppvec[12] = ppvec[13] = pg2 * p0;
         }
         else if (phdat == 12)
         {
            ppvec[14] = pg2 * pg0;
            ppvec[1] = p0 * p1;
            ppvec[5] = ppvec[6] = pg0 * p1;
            ppvec[7] = pg2 * p0;
         }
         else if ((phdat == 14) || (phdat == 13))
         {
            ppvec[14] = pg2 * pg1;
            ppvec[2] = ppvec[3] = p0 * p1;
            ppvec[5] = ppvec[6] = pg1 * p1;
            ppvec[10] = ppvec[11] = ppvec[12] = ppvec[13] = pg2 * p0;
         }
         else if (phdat == 15)
         {
            ppvec[14] = pg2 * pg2;
            ppvec[0] = p1;
            ppvec[1] = pg2;
            ppvec[2] = ppvec[3] = pg2;
            ppvec[4] = pg2;
            ppvec[8] = ppvec[9] = pg2;
            ppvec[5] = ppvec[6] = ppvec[7] = ppvec[10] = ppvec[11] = ppvec[12] =
               ppvec[13] = p1 * p1 * p1;
         }
      }
   }

   if ((ms == 9) || (ms == 15))
   {
      for (j = 0; j < ms - 1; ++j)
         ppvec[j] = ppvec[j] * (1.0 - errate) + errate * ppvec[ms - 1];
   }

   return;
}

void set_marg_ibd (int ms, double *ff)
{
   double cc, bb;
   bb = fkin;
   cc = bb * (1 - bb) / ((1 + bb) * (1 + bb + bb)); /* chi of the paper */

   if (ms == 2)
   {
      ff[0] = bb;
      ff[1] = 1.0 - bb;
   }

   else if (ms == 9)
   {
      ff[0] = 6.0 * bb * bb * cc / (1 - bb);
      ff[1] = bb * cc;
      ff[2] = ff[4] = 4.0 * ff[1];
      ff[3] = ff[5] = cc * (1 - bb);
      ff[6] = 2.0 * ff[1];
      ff[7] = 4.0 * ff[3];
      ff[8] = cc * (1 - bb) * (1 - bb) / bb;
   }

   else if (ms == 15)
   {
      ff[0] = 6.0 * bb * bb * cc / (1 - bb);
      ff[1] = ff[8] = ff[9] = bb * cc;
      ff[2] = ff[3] = ff[5] = ff[6] = 2.0 * ff[1];
      ff[4] = ff[7] = ff[10] = ff[11] = ff[12] = ff[13] = cc * (1 - bb);
      ff[14] = cc * (1 - bb) * (1 - bb) / bb;
   }
   return;
}

void set_transmat (int ms, double **totmat, double *ff, double dd)
/* set transition matrix at genetic distance dd, among ms states, at
   equilibrium probs ff and change rate ffr between markers at dist dd
   This is now called when needed within fwd/bk-compute        */
{
   int   i, k;
   double ffr;

   for (i = 0; i < ms; ++i)
      for (k = 0; k < ms; ++k)
         totmat[i][k] = 0.0;
   ffr = 1.0 - exp (-dd * ffrate);

   if (ms == 2)
   {
      for (i = 0; i < ms; ++i)
         totmat[i][i] = (1.0 - ffr);

      for (i = 0; i < ms; ++i)
         for (k = 0; k < ms; ++k)
            totmat[i][k] += ffr * ff[k];
   }
   else if ((ms == 15) || (ms == 9))
   {
      if (!tmatset)
      {
         set_tmat (ms);
         tmatset = 1;
         printf ("\n chk hh=%f, gg=%f, delta=%f", hh, gg, delta);
      }

      for (i = 0; i < ms; ++i)
      {
         for (k = 0; k < ms; ++k)
            totmat[i][k] = (1 - delta) * tmat[i][k] + delta * ff[k];
         totmat[i][i] -= delta;
      }

      for (i = 0; i < ms; ++i)
      {
         for (k = 0; k < ms; ++k)
            totmat[i][k] *= ffr;
         totmat[i][i] += 1.0;
      }
   }

   return;
}

void hap_fwd_compute (int *hd, int nd, int mph, int ms,
                      double **tmat2, double *allefrq,
                      double *fff, double *mdist, double **pprob, int igdx,
                      int *phasvec)
/* pprob[k, Sk] = P(Y0,...,Yk, Sk)  , 1 cluster, 2 states  */
/* NOTE now for 2-haps, 0 is IBD, 1 is NOT  */
/* NOTE now missing data, marker positions etc are accommodated
   in hap_set_penmat   

 * fff: marginal probs
 * pprob: matrix of forward probabilities
 * mdist: vector of inter-locus distances
 * pen2: penetrance vector (reset at each locus)
 * tmat2: transition matrix (reset at each locus)
 */
{
   int   i, ii, j, jj;
   double qn;
   double *pen2;
   int   phasint = 0;
   pen2 = malloc (ms * sizeof (double));

   printf ("\n starting fwd-compute for %d markers", nd);
   printf ("\n check phen %d, states %d", mph, ms);
   fflush (stdout);
   /* set the start values at locus 0 */

   /* if we are using the phase info, phasvec ought to be a non-null pointer */
   if (igdx == 2)
      phasint = phasvec[0];

   hap_set_penmat (allefrq[0], pen2, mph, ms, hd[0], 0.01, igdx, phasint);

   for (j = 0; j < ms; ++j)
      pprob[0][j] = fff[j] * pen2[j];

   for (i = 1; i < nd; ++i)
   {
      qn = 0.0;

      ii = skipmark * i;
      if (igdx == 2)
         phasint = phasvec[ii];
      hap_set_penmat (allefrq[ii], pen2, mph, ms, hd[i], 0.01, igdx, phasint);
      set_transmat (ms, tmat2, fff, mdist[i - 1]);

      for (j = 0; j < ms; ++j)
      {
         pprob[i][j] = 0;
         for (jj = 0; jj < ms; ++jj)
            pprob[i][j] += pprob[i - 1][jj] * tmat2[jj][j] * pen2[j];
         qn += pprob[i][j];
      }
      for (j = 0; j < ms; ++j)
      {
         pprob[i][j] = pprob[i][j] / qn;
      }
   }

   return;
}

void hap_bck_compute (int *hd, int nd, int mph, int ms, double **qibd,
                      double **tmat2, double *allefrq,
                      double *fff, double *mdist, double **pprob, int igdx,
                      int *phasvec)
/*  pprob[k][Sk] = P(Yk,... Yn | Sk)  from the end down--replacing the fwd */
/*  qibd = P(ibd given all the data)  
 *
 *  same variable names as in hap_fwd_compute */
{
   int   i, ii, j, jj;
   double *qi, qn;
   double *pen2;
   int   phasint = 0;
   pen2 = malloc (ms * sizeof (double));

   printf ("\n starting bck-compute for %d markers", nd);
   fflush (stdout);
   qi = malloc (ms * sizeof (double));
   alloc_chk (WHERE, qi);

   qn = 0.0;
   for (j = 0; j < ms; ++j)
      qn += pprob[nd - 1][j];
   /* when would qn ever not be == 1?  (normalized in hap_fwd_compute) */
   for (j = 0; j < ms; ++j)
      qibd[nd - 1][j] = pprob[nd - 1][j] / qn;

   ii = (nd - 1) * skipmark;
   if (igdx == 2)
      phasint = phasvec[ii];
   hap_set_penmat (allefrq[ii], pen2, mph, ms, hd[nd - 1], 0.01, igdx, phasint);
   for (j = 0; j < ms; ++j)
      pprob[nd - 1][j] = pen2[j];

   for (i = nd - 2; i >= 0; --i)
   {
      set_transmat (ms, tmat2, fff, mdist[i]);
      for (j = 0; j < ms; ++j)
      {
         qi[j] = 0;
         for (jj = 0; jj < ms; ++jj)
            qi[j] += pprob[i][j] * tmat2[j][jj] * pprob[i + 1][jj];
      }
      qn = 0.0;
      for (j = 0; j < ms; ++j)
         qn += qi[j];
      for (j = 0; j < ms; ++j)
         qibd[i][j] = qi[j] / qn;

      for (j = 0; j < ms; ++j)
      {
         qn = 0.0;
         pprob[i][j] = 0.0;
         ii = skipmark * i;
         if (igdx == 2)
            phasint = phasvec[ii];
         hap_set_penmat (allefrq[ii], pen2, mph, ms, hd[i], 0.01, igdx,
                         phasint);
         for (jj = 0; jj < ms; ++jj)
            pprob[i][j] += pen2[j] * tmat2[j][jj] * pprob[i + 1][jj];
         qn += pprob[i][j];
      }
      for (j = 0; j < ms; ++j)
         pprob[i][j] = pprob[i][j] / qn;
   }

   return;
}
