/* I decided to use signed long longs because although
 * an actual address is never negative, some addresses
 * get used in arithmetic with negative numbers and
 * it was going haywire with unsigned long longs.
 * long vs. long long probably doesn't matter but 
 * 1e8 base pairs isn't much smaller than the
 * signed long maximum, which is ~2e9 i believe --CGG */

typedef struct _Chromseg
{
   int   fgl;
   long long endpos;
} Chromseg;

/* why is this a separate struct from Chromseg--seems complicated? --CGG */
typedef struct _Chromseglist
{
   struct _Chromseg *current;
   struct _Chromseglist *next;
} Chromseglist;

typedef struct _Breaklist
{
    long long addr;
    struct _Breaklist * next;
} Breaklist;

typedef struct _Ibdlist
{
    long long start;
    long long end;
    struct _Ibdlist * next;
} Ibdlist;

typedef struct 
{
    int id;
    int motherid;
    int fatherid;
    int sex; /* 0 = female; 1 = male */
    Chromseglist * maternal;
    Chromseglist * paternal;
} Person;

typedef struct _Hashnode
{
    int fgl;
    int index;
    struct _Hashnode * next;
} Hashnode;

/* prototypes */
Chromseglist ** add_chrom_tail(Chromseglist ** tailaddr, int newfgl, long long newendpos);
Ibdlist ** add_ibd_tail(Ibdlist ** tailaddr, long long newstart, long long newend);
Chromseglist * recombine(Person x, Breaklist * newbreaks);
void weed_chrom(Chromseglist ** x);
Ibdlist * find_ibd(Chromseglist * first, Chromseglist * second);
void printbreaks(Chromseglist * x);
void printibd(Ibdlist * x, int id);
void printgeneration(Person * x, int length, int generation);
Breaklist * makebreaks (long long chrlength, double rate);
Person * initfounders(int n, long long chrlength, int startfgllab, int female);
Person read_person(FILE * fp);
Chromseglist * read_chrom(FILE * fp);
void free_chrom(Chromseglist * list);
void free_breaks(Breaklist * list);
void free_ibd(Ibdlist * list);
void free_generation(Person * list, int length);
void new_generation(Person * females, Person * males, \
        Person ** newfemales, Person ** newmales,\
        int nfemales, int nmales, int nkids, long long chrsize, double rate);

Hashnode ** add_hash_tail(Hashnode ** tailaddr, int newfgl, int newindex);
Hashnode ** make_hashtable();
int get_index(int fgl, Hashnode ** hashtable, int * nextindex);
char ** populate_snps(int nsnps, int npersons, FILE * fp);
Chromseglist * get_ibd(Chromseglist * first, Chromseglist * second);
int * ibd_fifteen(Person a, Person b, int nsnps, int * snp_pos);
int * ibd_nine(int * ibd_fifteen, int nsnps);
void print_table(FILE * outfile, Hashnode ** hashtable);
